package br.ufrj.aula15
import interp.Mem
import interp.Caixa
import br.ufrj.aula15.interp

object acao {
  type Acao[T] = Mem => (T, Mem)

  def id[T](v: T): Acao[T] = mem => (v, mem)
  
  def le(l: Int): Acao[Valor] = mem => mem.get(l) match {
    case Some(v) => (v, mem)
    case None => sys.error("endereço inválido: " + l)
  }
  
  def escreve(l: Int, v: Valor): Acao[Valor] = mem => (v, mem + (l -> v))  

  val next: Acao[Int] = for {
    NumV(free) <- le(0)
    _ <- escreve(0, NumV(free+1))
  } yield free.toInt
  
  def bind[T, U](a1: Acao[T], f: T => Acao[U]): Acao[U] = mem => {
    val (v, nmem) = a1(mem)
    val a2 = f(v)
    a2(nmem)  // => (v2: U, nmem2)
  }

  implicit class AcaoImp[T](val a: Acao[T]) extends AnyVal {
    def flatMap[U](f: T => Acao[U]): Acao[U] = bind(a, f)
    def map[U](f: T => U): Acao[U] = bind(a, (x: T) => id(f(x)))
    def withFilter(p: T => Boolean): Acao[T] = bind(a,
        (x: T) => if (p(x)) id(x) else sys.error("predicado falhou"))
  }
  
  def aloca(v: Valor): Acao[Valor] = for {
    nl <- next
    _ <- escreve(nl, v)
  } yield Caixa(nl)
  
  
}


