%{
package br.ufrj.aula8;
%}

%token FUN IF THEN ELSE END LET IN DEFINE TRUE FALSE NIL
%token<Double> NUM
%token<String> ID

%type<interp.Exp> exp
%type<interp.Prog> prog
%type<Set[interp.Fun1]> funs
%type<interp.Fun1> fun
%type<List[String]> params
%type<List[interp.Exp]> exps
%type<List[(String,interp.Exp)]> binds

%left '|'
%left '&'
%left '<'
%left '+' '-'
%left '*'
%left UNM '!'

%%

prog: funs exp   { res = Some(interp.Prog($1, $2)); }
    ;

funs:            { $$ = Set[interp.Fun1](); }
    | funs fun   { $$ = $1 + $2; }
    ;

fun : FUN ID '(' params ')' exp END  { $$ = interp.Fun1($2, $4, $6); }
    | FUN ID '(' ')' exp END { $$ = interp.Fun1($2, List(), $5); }
    ;

params: ID             { $$ = List($1); }
      | ID ',' params  { $$ = $1 :: $3; }
      ;

exp : NUM              { $$ = interp.Num($1); }
    | ID               { $$ = interp.Var($1); }
    | TRUE             { $$ = interp.True(); }
    | FALSE            { $$ = interp.False(); }
    | NIL              { $$ = interp.Nil(); }
    | IF exp THEN exp ELSE exp END { $$ = interp.If($2, $4, $6); }
    | ID '(' ')'       { $$ = interp.Ap1($1, List()); }
    | ID '(' exps ')'  { $$ = interp.Ap1($1, $3); }
    | LET ID '=' exp IN exp END { $$ = interp.Let($2, $4, $6); }
    | LET binds IN exp END { $$ = interp.LetM($2, $4); }
    | FUN '(' params ')' exp END { $$ = interp.Fun($3, $5); }
    | FUN '(' ')' exp END { $$ = interp.Fun(List(), $4); }
    | exp '+' exp      { $$ = interp.Soma($1, $3); }
    | exp '-' exp      { $$ = interp.Sub($1, $3); }
    | exp '*' exp      { $$ = interp.Mult($1, $3); }
    | exp '<' exp      { $$ = interp.Menor($1, $3); }
    | exp '&' exp      { $$ = interp.E($1, $3); }
    | exp '|' exp      { $$ = interp.Ou($1, $3); }
    | '!' exp          { $$ = interp.Nao($2); }
    | '-' exp %prec UNM  { $$ = interp.Menos($2); }
    | '(' exp ')'      { $$ = $2; }
    | '(' exp ')' '(' ')'  { $$ = interp.Ap($2, List()); }
    | '(' exp ')' '(' exps ')' { $$ = interp.Ap($2, $5); }
    ;

exps: exp             { $$ = List($1); }
    | exp ',' exps    { $$ = $1 :: $3; }
    ;

binds : ID '=' exp ',' ID '=' exp    { $$ = List(($1, $3), ($5, $7)); }
      | ID '=' exp ',' binds { $$ = ($1, $3) :: $5 }
      ;

%%

def yyerror(s:String) = println(s);
